#!/usr/bin/env python
# encoding: utf-8
"""
__init__.py

"""
__all__ = ['common', 'download', 'packager', 'tap', 'pdap', 'geogen']

# Set up the root logger

import logging
import sys
logging.basicConfig(format='%(levelname)s %(asctime)s (%(name)s): %(message)s',
                    level=logging.INFO, stream=sys.stdout, datefmt='%Y-%m-%d %H:%M:%S')





#!/usr/bin/python
"""pdap.py

Mark S. Bentley (mark@lunartech.org), 2021

A module to make PDAP queries of the PSA
"""

from astropy.io.votable import parse_single_table

psa_pdap_url = 'https://archives.esac.esa.int/psa/pdap'
import logging
import functools
import requests
import yaml
import pandas as pd
from io import BytesIO


log = logging.getLogger(__name__)
logging.getLogger("astropy").setLevel(logging.ERROR)


def exception(function):
    """
    A decorator that wraps the passed in function and handles
    exceptions raised by requests
    """
    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except requests.exceptions.HTTPError as e:
            log.error(e)
        except requests.exceptions.RequestException as e: 
            log.error(e)
    return wrapper


class Pdap:

    def __init__(self, pdap_url=psa_pdap_url):

        self.url = pdap_url

    def _url(self, path):
        """Helper function to append the path to the base URL"""
        
        return self.url + path

    @exception
    def get_datasets(self):
        """Retrieves meta-data for the set of datasets/bundles"""

        r = requests.get(
            self._url('/metadata'),
            params={
                'RETURN_TYPE': 'VOTABLE',
                'RESOURE_CLASS': 'DATA_SET'}) 
        r.raise_for_status()

        table = parse_single_table(BytesIO(r.content), pedantic=False)
        data = pd.DataFrame(table.array.data)
        
        return data

    @exception
    def get_products(self, dataset_id):
        """Queries the meta-data endpoint for products in the dataset ID
        given in the call"""

        r = requests.get(
            self._url('/metadata'),
            params={
                'RETURN_TYPE': 'VOTABLE',
                'RESOURCE_CLASS': 'PRODUCT',
                'DATA_SET_ID': dataset_id}) 
        r.raise_for_status()
        table = parse_single_table(BytesIO(r.content), pedantic=False)
        data = pd.DataFrame(table.array.data)

        # extract VIDs from the download url
        data['VID'] = data['PRODUCT.DATA_ACCESS_REFERENCE'].apply(lambda url: url.split('::')[-1])
        
        return data

    @exception
    def get_product(self, product_id):

        r = requests.get(
            self._url('/metadata'),
            params={
                'RETURN_TYPE': 'VOTABLE',
                'RESOURCE_CLASS': 'PRODUCT',
                'PRODUCT_ID': product_id}) 
        r.raise_for_status()
        table = parse_single_table(BytesIO(r.content), pedantic=False)
        data = pd.DataFrame(table.array.data)

        # extract VIDs from the download url
        data['VID'] = data['PRODUCT.DATA_ACCESS_REFERENCE'].apply(lambda url: url.split('::')[-1])
        
        return data.squeeze()

    @exception
    def get_files(self, dataset_id):

        r = requests.get(
            self._url('/files'),
            params={
                'RETURN_TYPE': 'VOTABLE',
                'RESOURCE_CLASS': 'PRODUCT',
                'DATA_SET_ID': dataset_id}) 
        r.raise_for_status()
        table = parse_single_table(BytesIO(r.content), pedantic=False)
        data = pd.DataFrame(table.array.data)

        return data


def latest_version(lid):
    """Uses PDAP to retrieve the highest VID for a given LID"""

    psa_pdap = Pdap()
    product = psa_pdap.get_product(lid)
    if len(product)==0:
        log.error('product with LID {:s} not found'.format(lid))
        return None
    else:
        version_list = product.VID.tolist()
        version_list.sort(key=lambda s: list(map(int, s.split('.'))))
        return version_list[-1]



#!/usr/bin/python
"""common.py

Mark S. Bentley (mark@lunartech.org), 2019

Common functions for the package

"""

pds_ns = 'http://pds.nasa.gov/pds4/pds/v1'
import logging
log = logging.getLogger(__name__)
import os

# def select_files(wildcard, directory='.', recursive=False):
#     """Create a file list from a directory and wildcard - recusively if
#     recursive=True"""

#     # recursive search
#     # result = [os.path.join(dp, f) for dp, dn, filenames in os.walk('.') for
#     # f in filenames if os.path.splitext(f)[1] == '.DAT']

#     if recursive:
#         selectfiles = locate(wildcard, directory)
#         filelist = [file for file in selectfiles]
#     else:
#         import glob
#         filelist = glob.glob(os.path.join(directory, wildcard))

#     filelist.sort()

#     return filelist


# def locate(pattern, root_path):
#     """Returns a generator using os.walk and fnmatch to recursively
#     match files with pattern under root_path"""

#     import fnmatch

#     for path, dirs, files in os.walk(os.path.abspath(root_path)):
#         for filename in fnmatch.filter(files, pattern):
#             yield os.path.join(path, filename)


def printtable(df, float_fmt=None):
    """Accepts a pd.DataFrame() prints a pretty-printed table, rendered with
    PrettyTable"""

    from prettytable import PrettyTable
    table = PrettyTable(list(df.columns))

    if float_fmt is not None:
        table.float_format = float_fmt

    for row in df.itertuples():
            table.add_row(row[1:])
    print(table)
    return



#!/usr/bin/python
"""tap.py

Mark S. Bentley (mark@lunartech.org), 2021

A module to make TAP queries of the PSA
"""

from astropy.io.votable import parse_single_table
from astroquery.utils.tap.core import Tap
import pandas as pd
from . import common
import time
from requests.exceptions import HTTPError

job_wait_time = 2 # seconds
job_wait_cycles = 10
psa_tap_url = 'https://archives.esac.esa.int/psa/epn-tap/tap/'
import logging
log = logging.getLogger(__name__)
logging.getLogger("astroquery").setLevel(logging.WARNING)


class PsaTap:

    def __init__(self, tap_url=psa_tap_url):
        """Establish a connection to the PSA TAP server"""
        self.tap = Tap(url=tap_url)


    def query(self, q, sync=True, dropna=True, verbose=False, job_wait_cycles=job_wait_cycles, job_wait_time=job_wait_time):
        """Make a simple query and return the data as a pandas DataFrame"""
        
        if sync:
            try:
                data = self.tap.launch_job(q, verbose=verbose).get_data()
            except ValueError as err:
                log.error('query error: {0}'.format(err))
                return None
            except HTTPError as err:
                log.error('http error: {0}'.format(err))
                return None
            data = data.to_pandas()
        else:
            job = self.tap.launch_job_async(q, verbose=verbose)
            for i in range(job_wait_cycles):
                time.sleep(job_wait_time)
                if job.is_finished():
                    break
            if not job.is_finished():
                log.error('asynchronous query did not complete')
                return None
            data = job.get_results().to_pandas()

        if 'time_min' in data.columns:
            data['time_min'] = pd.to_datetime(data['time_min'], origin='julian', unit='D') 

        if 'time_max' in data.columns:
            data['time_max'] = pd.to_datetime(data['time_max'], origin='julian', unit='D') 

        if dropna:
            data.dropna(inplace=True, axis=1, how='all')

        if len(data) == 2000:
            log.warn('results incomplete due to synchronous query limit - repeat with sync=false')

        return data# .squeeze()


def product_id_from_granule_uid(granule_uid):
    """Extracts ther PDS3 or PDS4 product ID from the granule_uid
    returned by EPN-TAP"""

    if granule_uid.startswith('urn:'): # it's PDS4
        product_id = granule_uid.split(':')[-3]
    else: # PDS3
        product_id = granule_uid.split(':')[-1]

    return product_id


def get_missions():
    tap = PsaTap()
    return tap.query('SELECT DISTINCT instrument_host_name from EPN_CORE').squeeze().tolist()

def get_instruments():
    tap = PsaTap()
    return tap.query('SELECT DISTINCT instrument_name from EPN_CORE').squeeze().tolist()

def get_collections(bundle_id):
     tap = PsaTap()
     return tap.query("select distinct granule_gid from epn_core where granule_gid like 'urn:esa:psa:{:s}:%%'".format(bundle_id))


def summarise_mission(mission_name, pretty=True):

    missions = get_missions()
    mission = [name for name in missions if name.lower()==mission_name.lower()]
    if len(mission)==0:
        log.error('mission name {:s} not found'.format(mission_name))
        return None
    else:
        tap = PsaTap()
        result = tap.query("SELECT instrument_name, count(*) FROM epn_core WHERE instrument_host_name='{:s}' GROUP BY instrument_name".format(mission[0]))
        if pretty:
            common.printtable(result)
        return result


def summarise_instrument(instrument_name, pretty=False):
 
    instruments = get_instruments()
    instrument = [name for name in instruments if name.lower()==instrument_name.lower()]
    if len(instrument)==0:
        log.error('instrument name {:s} not found'.format(instrument_name))
        return None
    else:
        tap = PsaTap()
        result = tap.query("SELECT processing_level, count(*) FROM epn_core WHERE instrument_name='{:s}' AND processing_level IS NOT NULL GROUP BY processing_level ORDER BY processing_level".format(instrument[0]))
        if pretty:
            common.printtable(result)
        return result




#!/usr/bin/python
"""geogen.py

Mark S. Bentley (mark@lunartech.org), 2019

A module of useful functions for working with geogen (PSA
geometry calculation package)

"""

import logging
log = logging.getLogger(__name__)
import json
import os


def generate_plf(config_file, files=None, directory='.', table=None, extras={}):
    """
    Generates a GEOGEN plf input file.

    pds4_utils.Database() is used to scrape meta-data according to the config_file.
    files= specifies the label file pattern (defaults to *.xml)
    directory= specified the root of the input files (and the output location)
    table= specifies the table name in case the input file is configured to
        produce more than one. Default (None) assumes only one table.
    extras = a dictionary which provides extra static key/value pairs to be added
        to every entry (e.g. product type or similar). If an identical value exists
        in the table and extras, extras has priority.
    """

    try:
        import pds4_utils
    except ModuleNotFoundError:
        log.error('pds4_utils module not available, please install before using psa_utils.geogen')
        return None

    # build a database of PDS4 meta-data using the specific config file
    dbase = pds4_utils.Database(files=files, directory=directory, config_file=config_file)

    # if the config file builds more than one table, we have to select this
    if table is None:
        tables = dbase.dbase.keys()
        if len(tables)>1:
            log.error('there are >1 tables in the specified configuration, specify with table=')
            return None
        t = list(tables)
        t = t[0]

    # get the column names from the "extra" meta-data
    table_cols = list(dbase.dbase[t].columns)

    # get the table back (merge the index with the scraped data)
    table = dbase.get_table(t)

    # re-map standard columns according to the expected input for GeoGen
    oldnames = ['bundle', 'product_id', 'start_time', 'stop_time']
    newnames = ['DATA_SET_ID', 'PRODUCT_ID', 'START_TIME', 'STOP_TIME']
    table.rename(columns={i:j for i,j in zip(oldnames, newnames)}, inplace=True)

    # the full set of columns is from both renamed and extra meta-data
    cols = newnames + table_cols 

    # add static column data from extras, if present
    for key in extras.keys():
        table[key] = extras[key]
        if key not in cols:
            cols.append(key)

    j = json.loads(table[cols].to_json(orient='records', date_format='iso'))
    products = {"products": j}
    json_file = os.path.join(directory, t + '.json')
    with open(json_file, 'w') as f:
        json.dump(products, f, indent=4)



#!/usr/bin/python
"""internal.py

Mark S. Bentley (mark@lunartech.org), 2019

A module to of utility functions likely only to be of interest to
PSA internal members (archive scientists etc.)

"""

from . import common
from . import packager
import pathlib
import logging
import copy
from lxml import etree
import shutil
import numpy as np
import pandas as pd
import datetime
import yaml
import os
from io import BytesIO

log = logging.getLogger(__name__)


# Class to produce systematic test ingestion products
proc_levels = {
    'raw': ('Raw', 'raw'),
    'par': ('Partially Processed', 'partially_processed'),
    'cal': ('Calibrated', 'calibrated'),
    'der': ('Derived', 'derived')
}

class Ingest_Test():
    """A class for generating test products from a label and data product
    template and a configuration file specifying the instrument-specific
    data"""

    def __init__(self, config_file='ingestion_test.yml', template_label='test_product.xml', output_dir='.', package=False):

        # read the configuration file
        self.read_config(config_file)
        if self.config is None:
            return

        self.load_template(template_label)
        if self.label is None:
            return

        self.generate_products(output_dir)

        log.info('Test products generated for {:d} bundles'.format(len(self.config)))

        if package:
            self.package(output_dir)


    def package(self, output_dir):
        """Produces one delivery package per bundle of the test data"""

        for bundle in self.config:
            prefix = self.config[bundle]['shortname']
            packager.Packager(input_dir=output_dir,  output_dir=output_dir, products='{:s}*.xml'.format(prefix))


    def read_config(self, config_file):
        """Read the YAML configuration file given by config_file and store in self.config"""

        try:
            f = open(config_file, 'r')
            self.config = yaml.load(f, Loader=yaml.BaseLoader)
        except FileNotFoundError:
            log.error('config file {:s} not found'.format(config_file))

        return


    def load_template(self, template_label):

        self.label = pathlib.Path(template_label)
        if not self.label.exists():
            log.error('could not open template file {:s}'.format(self.label.name))
            self.label = None

        self.tree = etree.parse(template_label)
        ns = self.tree.getroot().nsmap
        if None in ns and common.pds_ns == ns[None]:
            ns['pds'] = ns.pop(None)
        self.ns = ns


    def generate_products(self, output_dir):
        """Use the template_label and information in self.config to generate test products,
        one per configuration, and write the products to output_dir"""

        def get_product_id(bundle, level, sub_instr=None):

            if sub_instr is None:
                sub_instr = ''

            # product ID: {shortname}_{level}_sc_{sub_instr}_test_
            product_id = '{:s}_{:s}_sc_{:s}_test'.format(self.config[bundle]['shortname'], level, sub_instr)
            product_id = product_id.lower()

            return product_id


        def copy_data_file(root, output_dir):

            # copy and rename the template data product
            data_name = root.xpath("//pds:File_Area_Observational/pds:File/pds:file_name", namespaces=self.ns)[0].text
            data_file = pathlib.Path(os.path.join(self.label.parent.absolute(), data_name))
            if not data_file.exists():
                log.error('could not open data file file {:s}'.format(data_name))
                return None
            shutil.copy(data_file, os.path.join(output_dir, product_id + data_file.suffix))

            return data_file

        def update_template(root, bundle, product_id, level, data_file, sub_inst=None):

            mission, host, instrument = bundle.split('_')
            today = datetime.datetime.today()
            # update the template with the mission/instrument-relevant values

            root.xpath('//pds:Identification_Area/pds:logical_identifier', namespaces=self.ns)[0].text = 'urn:esa:psa:{:s}:data_{:s}:{:s}'.format(bundle, proc_levels[level][1], product_id)
            root.xpath('//pds:Identification_Area/pds:version_id', namespaces=self.ns)[0].text = '1.{:d}{:02d}'.format(today.year, today.month)
            root.xpath('//pds:Identification_Area/pds:Modification_History/pds:Modification_Detail/pds:modification_date', namespaces=self.ns)[0].text =today.strftime('%Y-%m-%d')
            root.xpath("//pds:Observing_System_Component[pds:type='Instrument']/pds:Internal_Reference/pds:lid_reference", namespaces=self.ns)[0].text = 'urn:esa:psa:context:instrument:{:s}.{:s}'.format(instrument, host)
            root.xpath("//pds:Observing_System_Component[pds:type='Instrument']/pds:name", namespaces=self.ns)[0].text = self.config[bundle]['fullname']
            root.xpath("//pds:Investigation_Area/pds:Internal_Reference[pds:reference_type='data_to_investigation']/pds:lid_reference", namespaces=self.ns)[0].text = 'urn:esa:psa:context:investigation:mission.{:s}'.format(mission)
            root.xpath("//pds:File_Area_Observational/pds:File/pds:file_name", namespaces=self.ns)[0].text = product_id + data_file.suffix
            root.xpath("//pds:Primary_Result_Summary/pds:processing_level", namespaces=self.ns)[0].text = proc_levels[level][0]
            if sub_inst:
            
                    # Need to insert new sub-instrument class, e.g.:
                    #
                    # <psa:Sub-Instrument>
                    #     <psa:identifier>STROFIO</psa:identifier>
                    #     <psa:name>STROFIO</psa:name>
                    #     <psa:type>Mass spectrometer</psa:type>
                    #     <psa:type>Gas analyser</psa:type>
                    # </psa:Sub-Instrument>

                    # find the Mission_Area
                    mission = root.xpath('//pds:Mission_Area', namespaces=self.ns)[0]

                    # create new elements, working down
                    sub = etree.SubElement(mission, '{{{:s}}}Sub-Instrument'.format(self.ns['psa']), nsmap=self.ns)
                    id = etree.SubElement(sub, '{{{:s}}}identifier'.format(self.ns['psa']), nsmap=self.ns)
                    id.text = sub_inst.upper()
                    name = etree.SubElement(sub, '{{{:s}}}name'.format(self.ns['psa']), nsmap=self.ns)
                    name.text = sub_inst
                    inst_type = etree.SubElement(sub, '{{{:s}}}type'.format(self.ns['psa']), nsmap=self.ns)
                    inst_type.text = self.config[bundle]['sub_instruments'][sub_inst]

                    # insert the new sub-instrument element into the mission element
                    mission.insert(idx, sub)

            return root

        def write_label(tree, output_dir, product_id):

            label_out = os.path.join(output_dir, product_id + '.xml')
            tree.write(label_out, xml_declaration=True, encoding=self.tree.docinfo.encoding) 

#########################

        for level in proc_levels:

            for bundle in self.config:

                if 'sub_instruments' in self.config[bundle].keys(): 
                    for idx, sub_inst in enumerate(self.config[bundle]['sub_instruments']):
                        tree = copy.deepcopy(self.tree)
                        root = tree.getroot()
                        product_id = get_product_id(bundle, level, sub_instr=sub_inst)
                        data_file = copy_data_file(root, output_dir)
                        if data_file is not None:
                            root = update_template(root, bundle, product_id, level, data_file, sub_inst)
                            write_label(tree, output_dir, product_id)

                else:
                    tree = copy.deepcopy(self.tree)
                    root = tree.getroot()
                    product_id = get_product_id(bundle, level, sub_instr=None)
                    data_file = copy_data_file(root, output_dir)
                    if data_file is not None:
                        root = update_template(root, bundle, product_id, level, data_file, None)
                        write_label(tree, output_dir, product_id)


def build_context_json(config_file, input_dir='.', output_dir='.', json_name='local_context_products.json', table='context_bundle'):
    """
    Generates a json file listing the name, type and LIDVID of all
    context files in input_dir. Generates a local context json file
    which can be used by the PDS validate tool and writes it to
    output_dir

    pds4_utils.Database() is used to scrape meta-data according to the config_file.
    """

    from pds4_utils import dbase
    import json

    context = []

    # build a database of context product meta-data
    dbase = dbase.Database(files='*.xml', directory=input_dir, config_file=config_file)
    table = dbase.get_table('context_bundle')

    if table is None:
        return None

    # create one entry for each unique LID
    for lid in table.lid.unique():
        c = {}
        # we only need the LATEST LID in the json file, since validate checks
        # for references to this or below
        product = table[table.lid==lid].sort_values(by='vid')
        product = product.iloc[-1]

        # the config file scrapes both instrument types, that from the IM
        # as type and that from the CTLI (newer IM versions) as ctli_type
        if product['type'] is None:
            prod_type = product['ctli_type']
        else:
            prod_type = product['type']
        
        if type(prod_type) == np.float:
            if pd.isna(prod_type):
                prod_type = ['N/A']

        # either way validate expects a list for type, even if it only has one value
        if type(prod_type) != list:
            t = [prod_type]
        else:
            t = prod_type

        # the json needs name, type, and lidvid
        c['type'] = t
        c['name'] = [product['name']]
        c['lidvid'] = "{:s}::{:s}".format(product.lid, product.vid)

        context.append(c)

    # the root of the json file needs to be Product_Context
    context = {"Product_Context": context}

    json_file = os.path.join(output_dir, json_name)

    # write to a file - indent=4 gives pretty printing
    with open(json_file, 'w') as f:
        json.dump(context, f, indent=4)
        log.info('written json file {:s}'.format(json_file))

    return 


def collection_summary(config_file, input_dir='.', output_dir='.', context_dir='.'):
    """
    collection_summary accesses meta-data in a collection label
    or referenced from it, to produce a set of summary information
    needed to register a DOI and/or create a Google Dataset
    Search landing page.
    """

    from pds4_utils import dbase

    collection_db = dbase.Database(
        files='collection_data*.xml', 
        config_file=config_file, 
        directory=input_dir, 
        recursive=True)

    collection_table = collection_db.get_table('collection')

    # strip carriage returns from the description
    collection_table.description = collection_table.description.apply(lambda desc: desc.replace('\n', ' '))

    # make a string list of keywords
    collection_table.keywords = collection_table.keywords.apply(lambda key: ', '.join(key))

    context_db = dbase.Database(
        files='*.xml', 
        config_file=config_file, 
        directory=context_dir, 
        recursive=True)

    context_table = context_db.get_table('context')
    # timeformatter = lambda x: pd.to_datetime(x).strftime('%Y-%m-%d %H:%M:%S.%f')

    for idx, entry in collection_table.iterrows():
        out_name = entry.bundle+'_'+entry.collection+'.html'
        out_file = os.path.join(output_dir, out_name)
        collection_cols = collection_db.dbase['collection'].columns.to_list()

        # tidy up some specific entries
        if entry.mission_lid is not None:
            mission_context = context_table[context_table.lid==entry.mission_lid]
            mission_descr = mission_context.mission_desc.squeeze()
            entry['mission_description'] = mission_descr.replace('\n', '')
        else:
            entry['mission_description'] = 'Mission description not found'
            log.warning('mission description not found for LID: {:s}'.format(entry.mission_lid))
        collection_cols.append('mission_description')
        collection_cols.remove('mission_lid')

        entry[collection_cols].to_frame().to_html(out_file, na_rep='')  
            # formatters={'start': timeformatter, 'stop': timeformatter})
        log.info('generated collection summary {:s}'.format(out_name))

    



#!/usr/bin/python
"""packager.py

Mark S. Bentley (mark@lunartech.org), 2021

Packages PDS4 products into a delivery package for ingestion into the PSA
"""

from . import common

import os
import sys
import pathlib
import datetime
import numpy as np
import pandas as pd
import shutil
import tarfile
import hashlib
from lxml import etree

import logging
log = logging.getLogger(__name__)

try:
    from pds4_utils import dbase
except ModuleNotFoundError:
    log.error('pds4_utils module not available, please install before using psa_utils.packager')

class Packager():

    def __init__(self, products='*.xml', input_dir='.', recursive=True, output_dir='.', template=None, 
        use_dir=False, clean=True, sendfrom=None, sendto=None):
        """Initialise the packager class. Accepts the following:

        products - file pattern to match labels (*.xml default)
        input_dir - the root directory for the labels (default=.)
        use_dir - uses the product directory structure for the archive
            (if false, a minimal structure will be adopted)
        clean - if True, removes the generated files are generating the tarball"""

        self.products = products
        self.input_dir = input_dir
        self.recursive = recursive
        self.output_dir = output_dir
        self.use_dir = use_dir
        self.index = None
        self.data_files = {}

        # sequentially run everything we need to build the delivery package
        self.get_products()              # index the specified products, get bundle, collection, etc.
        if not self.check_products():    # sanity checks - >1 bundle? etc.
            log.error('product checks failed, aborting')
            return None
        self.get_delivery_name(sendfrom, sendto)         # build the delivery package name

        self.package_dir = os.path.join(self.output_dir, self.delivery_name)
        os.makedirs(self.package_dir, exist_ok=True)

        self.build_paths()               # build delivery paths, according to use_dir
        self.create_transfer_manifest()  # create the transfer manifest .tab file
        self.create_checksum_manifest()  # create the checksum manifest
        self.create_label(template=template)              # flesh out the template PDS4 label
        self.create_package(clean=clean)            # copy files into correct structure and build tarball



    def get_products(self):
        """Obtain the list of products to be packaged and also list data products"""

        self.index = dbase.index_products(directory=self.input_dir, pattern=self.products, recursive=self.recursive)


    def check_products(self):
        """Perform basic sanity checks"""

        bad_products = []

        # check for products from multiple bundles
        if len(self.index.bundle.unique()) > 1:
            log.error('cannot package products from more than one bundle - aborting!')
            return False
        else:
            self.bundle = self.index.bundle.unique()[0]
            self.mission = self.bundle.split('_')[0]

        # check for duplicate products
        if self.index.duplicated(['lid','vid']).sum() > 0:
            log.error('duplicated product LIDVIDs in this package - aborting!')
            return False

        # check that all referenced data files are present
        for idx, product in self.index.iterrows():

            product_file = pathlib.Path(product.filename)
            root = etree.parse(product.filename).getroot()
                
            ns = root.nsmap
            if None in ns and common.pds_ns == ns[None]:
                ns['pds'] = ns.pop(None)

            product_type = root.xpath('name(/*)', namespaces=ns)
            if not product_type.startswith('Product_'):
                log.warn('XML file {:s} is not a PDS4 label, skipping'.format(product_file.name))
                bad_products.append(idx)
                continue

            # data_files =  root.xpath('//pds:File/pds:file_name', namespaces=ns)
            data_files =  root.xpath('//pds:file_name', namespaces=ns)
            for data_file in data_files:
                if not pathlib.Path(os.path.join(product_file.parent, data_file.text)).exists():
                    log.error('cannot find data file {:s} referenced in product {:s}, aborting!'.format(data_file, product_file.name))
                    return False
            self.data_files.update( {product.lid: [f.text for f in data_files]})

        if len(bad_products)>0:
            log.warn('{:d} products removed as invalid'.format(len(bad_products)))
            self.index.drop(bad_products, inplace=True)

        return True
        

    def get_delivery_name(self, sendfrom=None, sendto=None):

        self.delivery_time = datetime.datetime.now()

        if sendfrom is None:
            # get mission acronym from bundle
            # bundle = self.index.bundle.unique()[0]
            # source = bundle.split('_')[0]
            sendfrom = self.mission
        if sendto is None:
            sendto = 'psa'


        self.delivery_name = '{:s}{:s}-pds4-PI-01-{:s}-{:s}'.format(sendfrom, sendto, self.bundle, self.delivery_time.strftime('%Y%m%dT%H%M%S'))


    def build_paths(self):

        self.index['path'] = None

        for idx, product in self.index.iterrows():

            label = pathlib.Path(product.filename)

            if not self.use_dir:
                # simply use the bundle/collection root
                self.index.path.loc[idx] = os.path.join(product.bundle, product.collection, label.name)
            else:
                # use the path relative to the input directory
                self.index.path.loc[idx] = os.path.join(product.bundle, os.path.relpath(label, start=self.input_dir))

        return



    def create_transfer_manifest(self):

        # create tab seaparated files
        manifest_file = self.delivery_name + '-transfer_manifest.tab'
        self.manifest_file = os.path.join(self.package_dir, manifest_file)

        path_len = self.index.path.str.len().max()
        lid_len = self.index.lid.str.len().max()
        self.transfer_fields = {
            'lid_start': 1,
            'lid_len': lid_len,
            'path_start': lid_len + 1,
            'path_len': path_len}

        # pandas to_string does weird things and pads strings, so using numpy instead
        np.savetxt(self.manifest_file, self.index[['lid','path']].values, fmt='%-{:d}s%-{:d}s'.format(lid_len+1, path_len), newline='\r\n')

        self.transfer_records =  len(self.index)

        return


    def create_checksum_manifest(self, template=None):

        checksum = pd.DataFrame([], columns=['checksum', 'filepath'])
        check = []
        filepath = []

        for idx, product in self.index.iterrows():
            # add the label checksum
            check.append(self.md5_hash(product.filename))
            filepath.append(product.path)

            # add the data file checksums
            for data_file in self.data_files[product.lid]:
                data_file_absolute = os.path.join(pathlib.Path(product.filename).parent, data_file)
                check.append(self.md5_hash(data_file_absolute))
                filepath.append(os.path.join(pathlib.Path(product.path).parent, data_file))

        checksum['checksum'] = check
        checksum['filepath'] = filepath

        checksum_file = self.delivery_name + '-checksum_manifest.tab'
        self.checksum_file = os.path.join(self.package_dir, checksum_file)
        checksum.to_csv(self.checksum_file, line_terminator='\r\n', sep='\t', header=False, index=False)
        
        self.checksum_records = len(checksum)

        return


    def create_label(self, template=None):

        if template is None:
            # if no template is given, look for defaults
            template_path = os.path.dirname(__file__)
            template_file = os.path.join(template_path, 'templates', 'product_delivery_template.xml')
        else:
            template_file = template
        
        if not pathlib.Path(template_file).exists():
            log.error('could not open template file {:s}'.format(template.name))
            return None
        else:
            tree = etree.parse(template_file)
            root = tree.getroot()
            ns = root.nsmap
            if None in ns and common.pds_ns == ns[None]:
                ns['pds'] = ns.pop(None)

        root.xpath('/pds:Product_AIP/pds:Identification_Area/pds:logical_identifier', namespaces=ns)[0].text = 'urn:esa:psa:{:s}:data_delivery:{:s}'.format(self.mission, self.delivery_name.lower())
        root.xpath('/pds:Product_AIP/pds:Information_Package_Component/pds:Internal_Reference/pds:lid_reference', namespaces=ns)[0].text = 'urn:esa:psa:{:s}:{:s}'.format(self.mission, self.bundle)
        
        cf = pathlib.Path(self.checksum_file)
        root.xpath('/pds:Product_AIP/pds:Information_Package_Component/pds:File_Area_Checksum_Manifest/pds:File/pds:file_name', namespaces=ns)[0].text  = self.delivery_name + '-checksum_manifest.tab'
        root.xpath('/pds:Product_AIP/pds:Information_Package_Component/pds:File_Area_Checksum_Manifest/pds:File/pds:creation_date_time', namespaces=ns)[0].text = self.delivery_time.strftime('%Y-%m-%dT%H:%M:%SZ')
        root.xpath('/pds:Product_AIP/pds:Information_Package_Component/pds:File_Area_Checksum_Manifest/pds:File/pds:file_size', namespaces=ns)[0].text = str(cf.stat().st_size)
        root.xpath('/pds:Product_AIP/pds:Information_Package_Component/pds:File_Area_Checksum_Manifest/pds:File/pds:records', namespaces=ns)[0].text = str(self.checksum_records)
        root.xpath('/pds:Product_AIP/pds:Information_Package_Component/pds:File_Area_Checksum_Manifest/pds:File/pds:md5_checksum', namespaces=ns)[0].text = self.md5_hash(self.checksum_file)

        tf = pathlib.Path(self.manifest_file)
        root.xpath('/pds:Product_AIP/pds:Information_Package_Component/pds:File_Area_Transfer_Manifest/pds:File/pds:file_name', namespaces=ns)[0].text = self.delivery_name + '-transfer_manifest.tab'
        root.xpath('/pds:Product_AIP/pds:Information_Package_Component/pds:File_Area_Transfer_Manifest/pds:File/pds:creation_date_time', namespaces=ns)[0].text = self.delivery_time.strftime('%Y-%m-%dT%H:%M:%SZ')
        root.xpath('/pds:Product_AIP/pds:Information_Package_Component/pds:File_Area_Transfer_Manifest/pds:File/pds:file_size', namespaces=ns)[0].text =str(tf.stat().st_size)
        root.xpath('/pds:Product_AIP/pds:Information_Package_Component/pds:File_Area_Transfer_Manifest/pds:File/pds:records', namespaces=ns)[0].text = str(self.transfer_records)

        root.xpath('/pds:Product_AIP/pds:Information_Package_Component/pds:File_Area_Transfer_Manifest/pds:Transfer_Manifest/pds:Record_Character/pds:Field_Character[1]/pds:field_length', namespaces=ns)[0].text = str(self.transfer_fields['lid_len'])
        root.xpath('/pds:Product_AIP/pds:Information_Package_Component/pds:File_Area_Transfer_Manifest/pds:Transfer_Manifest/pds:Record_Character/pds:Field_Character[2]/pds:field_location', namespaces=ns)[0].text = str(self.transfer_fields['path_start'])
        root.xpath('/pds:Product_AIP/pds:Information_Package_Component/pds:File_Area_Transfer_Manifest/pds:Transfer_Manifest/pds:Record_Character/pds:Field_Character[2]/pds:field_length', namespaces=ns)[0].text = str(self.transfer_fields['path_len'])
        root.xpath('/pds:Product_AIP/pds:Information_Package_Component/pds:File_Area_Transfer_Manifest/pds:Transfer_Manifest/pds:Record_Character[1]/pds:record_length', namespaces=ns)[0].text = str(self.transfer_fields['path_start'] + self.transfer_fields['path_len'] + 2)

        root.xpath('/pds:Product_AIP/pds:Information_Package_Component/pds:File_Area_Transfer_Manifest/pds:File/pds:md5_checksum', namespaces=ns)[0].text = self.md5_hash(self.manifest_file)
        root.xpath('/pds:Product_AIP/pds:Information_Package_Component/pds:File_Area_Transfer_Manifest/pds:Transfer_Manifest/pds:records', namespaces=ns)[0].text = str(self.transfer_records)
        root.xpath('/pds:Product_AIP/pds:Archival_Information_Package/pds:description', namespaces=ns)[0].text = 'Generated by bepicolombo.psa_utils.Packager'

        # write out the modified label
        label_file = os.path.join(self.package_dir, self.delivery_name + '.xml')
        tree.write(label_file, xml_declaration=True, encoding=tree.docinfo.encoding) 

        return

    def create_package(self, clean):

        # create a directory
        product_dir = os.path.join(self.package_dir, self.bundle)
        os.makedirs(product_dir, exist_ok=True)

        for idx, product in self.index.iterrows():

            # create the directory structure
            path = os.path.join(self.package_dir, pathlib.Path(product.path).parent)

            try:
                os.makedirs(path, exist_ok=True)
            except OSError:
                log.error ("creation of the directory %s failed" % path)
                return None

            # copy the label
            shutil.copy(product.filename, path)

            # copy the data files referenced
            for data_file in self.data_files[product.lid]:
                data_file_absolute = os.path.join(pathlib.Path(product.filename).parent, data_file)
                shutil.copy(data_file_absolute, path)

        tarball = os.path.join(self.output_dir, self.delivery_name + '.tar.gz')
        with tarfile.open(tarball, "w:gz") as tar:
            tar.add(self.package_dir, arcname=os.path.basename(self.package_dir))

        if clean:
            shutil.rmtree(self.package_dir)

        return


    def md5_hash(self, filename):

        BLOCKSIZE = 65536
        hasher = hashlib.md5()
        with open(filename, 'rb') as f:
            buf = f.read(BLOCKSIZE)
            while len(buf) > 0:
                hasher.update(buf)
                buf = f.read(BLOCKSIZE)
        return hasher.hexdigest()



#!/usr/bin/python
"""download.py

Mark S. Bentley (mark@lunartech.org), 2021

A module to download public PSA products
"""

import os
import requests
import re
import pathlib
import pandas as pd
from lxml import etree

from . import pdap
from . import tap

import logging
log = logging.getLogger(__name__)


def download_label_by_granule_uid(granule_uid, output_dir='.'):
    """
    Accepts a granule_uid returned from EPN-TAP (essentially a
    LIDVID) and downloads it.
    """

    label_url = get_label_urls([granule_uid])
    filename = os.path.basename(label_url)
    if label_url is None:
        log.error('could not retrieve label URL')
        return None
    local_filename = download_file(label_url, output_dir, filename)
    return local_filename



def download_file(url, output_dir='.', output_file=None):
    """
    Downloads the file specified by url to the local directory specified
    by output_dir.

    If output_file is set, this wil be used as the output filename.
    If output_file is None, an attempt will be made to get the filename
    from the content-disposition header.
    """
    
    path = pathlib.Path(output_dir)
    path.mkdir(parents=True, exist_ok=True)

    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        if output_file is None:
            filename = get_filename_from_cd(r.headers.get('content-disposition'))
        else:
            filename = output_file
        local_filename = os.path.join(output_dir, filename)
        with open(local_filename, 'wb') as f:
            for chunk in r.iter_content(chunk_size=8192): 
                f.write(chunk)
    log.debug('downloaded file {:s}'.format(filename))
    return local_filename


def get_filename_from_cd(cd):
    """
    Get filename from content-disposition
    """
    if not cd:
        return None
    fname = re.findall('filename=(.+)', cd)
    if len(fname) == 0:
        return None

    return fname[0].strip('\"')


def download_by_lid(lid, output_dir='.', unzip=True, tidy=True):

    query = "select access_url, granule_uid from epn_core where granule_uid like '%%{:s}%%'".format(lid)
    files = download_by_query(query, output_dir, tidy)
    return files



def download_by_query(query, output_dir='.', unzip=True, tidy=True):
    """
    Runs a query against the PSA's EPN-TAP interface. Any products which match,
    and are public (have a download URL) will be downloaded and the zips placed
    into output_dir. If unzip=True they will be unzipped into output_dir and
    if tidy=True the zips will be removed after use
    """

    from zipfile import ZipFile

    psa_tap = tap.PsaTap()

    if tidy and not unzip:
        log.warning('cannot remove source files without decompressiong - setting tidy=False')
        tidy = False

    files = []

    products = psa_tap.query(query)
    if products is None:
        log.error('no products matching query')
    if ('granule_uid' not in products.columns) or ('access_url' not in products.columns):
        log.error('queries have to return granule_uid and access_url for product download')
        raise ValueError
    for idx, product in products.iterrows():
        product_id = tap.product_id_from_granule_uid(product.granule_uid)
        if product.access_url == '':
            log.warning('skipping proprietary product {:s}'.format(product_id))
            continue
        else:
            log.info('downloading product {:s}'.format(product_id))
            try:
                local_file = download_file(product.access_url, output_dir=output_dir)
                if not unzip:
                    files.append(local_file)
            except:
                log.error('failure to download {:s}, skipping'.format(product_id))
                continue
                
        if unzip:
            with ZipFile(local_file, 'r') as zipObj:
                zipObj.extractall(output_dir)
                filelist = zipObj.namelist()
                for f in filelist:
                    files.append(os.path.join(output_dir, f))

        if tidy:
            os.remove(local_file)

        files = list(set(files))

    return files

def download_labels_by_query(query, output_dir='.'):

    psa_tap = tap.PsaTap()
    products = psa_tap.query(query)
    if products is not None:
        download_labels(products, output_dir)

    return


def download_labels(epn_tap_df, output_dir='.'):
    """Accepts a DataFrame as returned by psa_tap.query, uses 
    get_label_urls to add applicable URLs to the DataFrame and
    then downloads each label to output_dir
    """
    epn_tap_df = get_label_urls(epn_tap_df)
    for idx, url in epn_tap_df.label_url.iteritems():
        if url is None: # skip PDS3 or proprietary labels
            continue
        filename = os.path.basename(url)
        download_file(url, output_dir='.', output_file=filename)

    return

def get_label_urls(epn_tap_df):
    """Accepts a DataFrame as returned by psa_tap.query, filters for
    PDS4 products, and finds the unique bundles.
    
    For each bundle it retrieves the corresponding file list from PDAP
    and then adds the download URL for the label, returning the df.

    Note that this will only work for detached labels!
    """

    epn_tap_df['label_url'] = None

    epn_tap_df['pds4'] = epn_tap_df.granule_uid.apply(
            lambda uid: True if uid.startswith('urn:') else False)

    epn_tap_df = epn_tap_df[epn_tap_df.access_url != ''] # remove proprietary entries

    # add the bundle/dataset ID 
    epn_tap_df['bundle'] = None

    epn_tap_df.loc[epn_tap_df.pds4, 'bundle'] = epn_tap_df[epn_tap_df.pds4].granule_gid.apply(lambda uid: ':'.join(uid.split(':')[0:4]))
    epn_tap_df.loc[~epn_tap_df.pds4, 'bundle'] = epn_tap_df[~epn_tap_df.pds4].granule_gid.apply(lambda uid: uid.split(':')[0])
    
    psa_pdap = pdap.Pdap()

    for bundle in epn_tap_df.bundle.unique():

        log.debug('querying files in bundle {:s}'.format(bundle))
        files = psa_pdap.get_files(bundle)

        # loop through products from the query that are in this bundle
        bundle_products = epn_tap_df[epn_tap_df.bundle==bundle]
        for idx, product in bundle_products.iterrows():
                product_id = tap.product_id_from_granule_uid(product.granule_uid)
                product_files = files[files.ProductId==product_id].copy()
                if len(product_files)==0:
                    log.error('could not find product {:s} in bundle {:s}'.format(product_id, bundle))
                    continue

                # get the extensions to identify labels
                product_files['ext'] = product_files.Filename.apply(lambda x: os.path.splitext(x)[-1])

                # note that PDS3 attached labels will be skipped!
                label_files = product_files[product_files.ext.str.lower().isin(['.xml','.lbl'])]
                if len(label_files)==0:
                    log.error('could not find label for product {:s}'.format(product_id))
                    return None
                elif len(label_files)>1:
                    log.error('more than one file found for product {:s}'.format(product_id))
                    return None
                else:
                    epn_tap_df.loc[idx, 'label_url'] = label_files.Reference.squeeze()

    return epn_tap_df


def read_label_by_url(label_url):
    """Parses a PDS4 label into memory given its URL"""

    try:
        response = requests.get(label_url)
        root = etree.fromstring(bytes(response.text, encoding='utf-8'))
    except:
        log.error('problem retrieving label')
        # todo check exceptions raised here

    return root


def read_label_by_lid(lid):

    query = "select access_url, granule_gid, granule_uid from epn_core where granule_uid like '%%{:s}%%'".format(lid)
    psa_tap = tap.PsaTap()
    products = psa_tap.query(query)
    products = get_label_urls(products)
    root = read_label_by_url(products.label_url.squeeze())

    return root


